/*
 * udpadmin -- Julius と Asterisk を連携するためのプラグイン
 *
 * Copyright (C) 2009, iWeave, Inc.
 *
 * Yasuhiro Konishi <konishi@iweave.jp>
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 *
 * なお、本ソースは、「大語彙連続音声認識エンジン Julius」に付属している
 * plugin のサンプルソースをベースに作成しました。
 * もともとのソールのライセンスは、このディレクトリにある Julius-LICENSE.txt となります。
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/ioctl.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <julius/juliuslib.h>
#include "plugin_defs.h"

static int freq;            /* given sampling frequency */

static int sock;
static int udpport = 12345;
static struct sockaddr_in addr;
static struct sockaddr_in from;
static socklen_t from_len = 0;

Recog *recog = NULL;

/* return plugin information */
int
get_plugin_info(int opcode, char *buf, int buflen)
{
	switch(opcode) {
	case 0:
		strncpy(buf, "UDP adin plugin", buflen);
		break;
	}
	return 0;
}

/* return argument string for "-udpadmin" */
void
adin_get_optname(char *buf, int buflen)
{
	strncpy(buf, "udpadin", buflen);
}

static boolean
opt_udpport(Jconf *jconf, char *arg[], int argnum)
{
	udpport = atoi(arg[0]);
	fprintf(stderr, "udpport %d\n", udpport);
	return TRUE;
}

int
initialize()
{
	j_add_option("-udpport", 1, 1, "udp port", opt_udpport);
	return 0;
}

/* return property of this adin input */
int
adin_get_configuration(int opcode)
{
	switch(opcode) {
	case 0:       /* enable real-time processing of 1st pass by default? */
		return 1;   /* yes */
	case 1:       /* enable frontend voice detection by level/zc by default? */
		return 1;   /* yes */
	case 2:       /* input module threading is needed or not, if supported? */
		return 1;   /* yes, needed*/
	}
	return 0;
}

/* standby, will be called once at startup */
boolean
adin_standby(ADIn *a, int sfreq, void *arg)
{
	freq = sfreq;  /* just store required sampling frequency to local */
	return TRUE;
}

/* open device */
boolean
adin_open()
{
	sock = socket(AF_INET, SOCK_DGRAM, 0);

	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(udpport);
	addr.sin_addr.s_addr = INADDR_ANY;

	bind(sock, (struct sockaddr *)&addr, sizeof(addr));

	/* ブロックしない */
	int val = 1;
	ioctl(sock, FIONBIO, &val);

	memset(&from, 0, sizeof(from));

	return TRUE;
}

#define EAGAIN_TIMEOUT 20
static int eagain_count = EAGAIN_TIMEOUT;

/* read samples */
int
adin_read(SP16 *buf, int sampnum)
{
	int cnt;

	/* パケットを受信する */
	from_len = sizeof(from);
	cnt = recvfrom(sock, (void*)buf, sizeof(SP16) * sampnum, 0, (struct sockaddr *)&from, &from_len);

	if (cnt < 0) {
		/* 読み込みエラー時の処理 */
		if (errno == EAGAIN) {
			/* パケットを受信できなかった場合 */
			eagain_count++;
			usleep(20 * 1000);
			/* 暫定対応：一定期間パケットを受信できない場合は、切断と見なす */
			return (eagain_count == EAGAIN_TIMEOUT) ? ADIN_EOF : 0;
		}
		/* 何かのエラー */
		fprintf(stderr, "Error: adin_udp: failed to read samples\n");
		return ADIN_ERROR;
	}

	/* パケットを受信できた */
	eagain_count = 0;
	cnt /= sizeof(SP16);
	return cnt;
}

/* close device */
boolean
adin_close()
{
	close(sock);
	return TRUE;
}

/* 認識結果を返す */
void
result_best_str(char *result_str)
{
	int cnt;
	if (eagain_count < EAGAIN_TIMEOUT) {
		if (result_str == NULL) {
			/* 認識失敗時 */
			cnt = sendto(sock, (void*)"#", 1, 0, (struct sockaddr *)&from, from_len);
			fprintf(stderr, "\n== recognize: FAILURE\n");
		} else {
			/* 認識成功時 */
			cnt = sendto(sock, (void*)result_str, strlen(result_str), 0, (struct sockaddr *)&from, from_len);
			fprintf(stderr, "\n-- recognized: '%s'\n", result_str);
		}
	} else {
		fprintf(stderr, "\n** recognized: '%s'\n", result_str);
	}
}
